# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains all tests for the Visualize() class. Note that the Visualize class is NOT case dependent.
Therefore, an arbitrary outcome dictionary is used in these unit tests.

The following functions are skipped: _table_styler, _graph_styler, _create_table (returns styler class),
The following functions are partly tested: _create_barchart (only errors), create_visuals (only errors)
"""
import pytest
import numpy as np
import pandas as pd
from params import OUTPUT_DICT_GENERIC
from vlinder.visualize import Visualize, VisualizationError


@pytest.fixture(name="test_outcomes")
def fixture_visualize_outcomes():
    """
    This fixture initialises a Visualize class.
    :return: a Visualize class for a generic outcome dictionary
    """
    return Visualize(OUTPUT_DICT_GENERIC, 2 * 3 * 3)


def test_validate_kwargs(test_outcomes):
    """
    This function tests _validate_kwargs to return None when valid **kwargs are provided
    :param test_outcomes: a Visualize class
    """
    result = test_outcomes._validate_kwargs(scenario="A", decision_makers_option="B", stacked=False)
    expected_result = None
    assert result == expected_result


def test_validate_kwargs_error(test_outcomes):
    """
    This function tests that _validate_kwargs to raise a VisualizationError when an unknown argument is provided.
    :param test_outcomes: a Visualize class
    """
    with pytest.raises(VisualizationError) as visualization_error:
        test_outcomes._validate_kwargs(this_does_not_exist="abc")
    expected_result = "Visualization Error: Invalid argument 'this_does_not_exist'"
    assert str(visualization_error.value) == expected_result


@pytest.mark.parametrize("target_key, expected_result", [("SCEN A", 1), ("AA", None), ("val2", 3)])
def test_find_dimension_level(test_outcomes, target_key, expected_result):
    """
    This function tests _find_dimension_level to return the correct dimension level.
    :param test_outcomes: a Visualize class
    :param target_key: the key of interest
    :param expected_outcome: the expected dimension level
    """
    result = test_outcomes._find_dimension_level(OUTPUT_DICT_GENERIC, target_key)
    assert result == expected_result


def test_str_snake_case_to_text(test_outcomes):
    """
    This function tests _str_snake_case_to_text to return a reformatted string.
    :param test_outcomes: a Visualize class
    """
    result = test_outcomes._str_snake_case_to_text("string_to_be_adjusted")
    expected_result = "string to be adjusted"
    assert result == expected_result


@pytest.mark.parametrize(
    "title_list, expected_result",
    [
        (
            ["this is a very long title name that needs to become shorter", "a", "b", "c", "d", "e"],
            ["this is a ve..", "a", "b", "c", "d", "e"],
        ),
        (["short", "shorter", "shorts"], ["short", "shorter", "shorts"]),
        (["very long but also only one so should be fine"], ["very long but also only one so should be fine"]),
    ],
)
def test_truncate_title_list(test_outcomes, title_list, expected_result):
    """
    This function tests _truncate_title_list to return a truncated list of title strings.
    :param test_outcomes: a Visualize class
    :param title_list: a list with strings
    :param expected_result: the expected truncated list
    """
    result = test_outcomes._truncate_title_list(title_list)
    assert result == expected_result


@pytest.mark.parametrize(
    "key_data, expected_result",
    [
        (
            "val1",
            pd.DataFrame(
                {
                    "scenario": ["SCEN A"] * 9 + ["SCEN B"] * 9,
                    "decision_makers_option": ["DMO 1"] * 3
                    + ["DMO 2"] * 3
                    + ["DMO 3"] * 3
                    + ["DMO 1"] * 3
                    + ["DMO 2"] * 3
                    + ["DMO 3"] * 3,
                    "val1": ["KO1", "KO2", "KO3"] * 3 + ["KO1", "KO2", "KO3"] * 3,
                    "value": [
                        2.0,
                        4.0,
                        3.0,
                        3.0,
                        2.0,
                        4.0,
                        5.0,
                        2.5,
                        3.5,
                        2.5,
                        3.0,
                        2.0,
                        4.0,
                        1.5,
                        2.5,
                        1.0,
                        3.0,
                        2.5,
                    ],
                }
            ),
        ),
        (
            "val3",
            pd.DataFrame(
                {
                    "scenario": ["SCEN A", "SCEN A", "SCEN A", "SCEN B", "SCEN B", "SCEN B"],
                    "decision_makers_option": ["DMO 1", "DMO 2", "DMO 3", "DMO 1", "DMO 2", "DMO 3"],
                    "value": [5.0, 2.0, 4.0, 1.5, 2.0, 3.5],
                }
            ),
        ),
    ],
)
def test_format_data_for_visual(test_outcomes, key_data, expected_result):
    """
    This function tests _format_data_for_visual to return a properly formatted dataframe.
    :param test_outcomes: a Visualize class
    :param key_data: key name of the requested values
    :param expected_result: expected dataframe
    """
    result = test_outcomes._format_data_for_visual(key_data)
    assert result.to_dict() == expected_result.to_dict()


@pytest.mark.parametrize(
    "drop_used, expected_result",
    [
        (True, pd.DataFrame({"Col2": ["X", "X", "Y"]})),
        (False, pd.DataFrame({"Col1": ["A", "A", "A"], "Col2": ["X", "X", "Y"], "Col3": [3, 3, 3]})),
    ],
)
def test_apply_filters(test_outcomes, drop_used, expected_result):
    """
    This function tests _apply_filters to return a correctly filtered dataframe.
    :param test_outcomes: a Visualize class
    """
    np.random.seed(42)
    test_data = pd.DataFrame(
        {
            "Col1": np.random.choice(["A", "B", "C"], 20),
            "Col2": np.random.choice(["X", "Y", "Z"], 20),
            "Col3": np.random.choice([1, 2, 3], 20),
        }
    )
    result = test_outcomes._apply_filters(test_data, drop_used, Col1="A", Col3=3, other_arg="abc")
    result_df = result[0].reset_index(drop=True)
    assert result_df.to_dict() == expected_result.to_dict()


def test_apply_filters_error(test_outcomes):
    """
    This function tests _apply_filters to raise a VisualizationError if the filter removes all data.
    :param test_outcomes: a Visualize class
    """
    test_data = pd.DataFrame({"ColA": ["A", "A"], "ColB": [1, 2]})
    expected_error = "Visualization Error: No data for given selection. Are your arguments correct?"
    with pytest.raises(VisualizationError) as visualization_error:
        test_outcomes._apply_filters(test_data, False, ColA="B")
    assert str(visualization_error.value) == expected_error


def test_create_barchart_error(test_outcomes):
    """
    This function tests _create_barchart to raise an error when too many dimensions are provided
    :param test_outcomes: a Visualize class
    """
    with pytest.raises(VisualizationError) as visualization_error:
        test_outcomes._create_barchart("val1")
    expected_error = "Visualization Error: Too many dimensions (3). Please specify a scenario"
    assert str(visualization_error.value) == expected_error


@pytest.mark.parametrize(
    "visual_request, key, expected_error",
    [
        ("non_existing_plot", "key_outputs", "'non_existing_plot' is not a valid chart type"),
        ("barchart", "non_existing_value", "'non_existing_value' is not a valid option"),
    ],
)
def test_create_visual_errors(test_outcomes, visual_request, key, expected_error):
    """
    This function tests create_visual to raise an error for invalid request
    :param test_outcomes: a Visualize class
    :param visual_request: requested visual
    :param key: key name of requested value
    :expected_error: expected raised error
    """
    with pytest.raises(VisualizationError) as visualization_error:
        test_outcomes.create_visual(visual_request, key)
    assert str(visualization_error.value) == f"Visualization Error: {expected_error}"



"""
This module contains general helper functions
"""


def round_all_dict_values(my_dict: dict, digits: int = 2) -> dict:
    """
    This function rounds all values in a dictionary to the number of digits specified. Works for nested dictionaries.
    :param my_dict: (nested) dictionary
    :param digits: number of digits of the rounded values
    :return: dictionary with rounded values
    """
    rounded_dict = {}
    for key, value in my_dict.items():
        if isinstance(value, dict):
            rounded_dict[key] = round_all_dict_values(value, digits)
        elif isinstance(value, list):
            rounded_dict[key] = [round(val, digits) for val in value]
        else:
            rounded_dict[key] = round(value, digits)
    return rounded_dict


def get_values_from_target(dictionary: dict, target: str) -> list:
    """
    This function return all (nested) values for a given dictionary and target key
    :param dictionary: dictionary where values need to be subtracted from
    :param target: target key
    :return: list of values
    """
    values = []

    for key, value in dictionary.items():
        if key == target:
            values.append(value)
        elif isinstance(value, dict):
            values.extend(get_values_from_target(value, target))

    return values


def number_formatter(number: float) -> str:
    """
    This function transforms a value into a rounded value: thousands --> K, millions --> M
    :param number: input value
    :return: formatted string, ending with M, K, or two digits (depending on number size)
    """
    if abs(number) >= 1e6:
        return f"{number/1e6:.1f}M"
    if abs(number) >= 1e3:
        return f"{number/1e3:.1f}K"
    return f"{number:.2f}"


def check_numeric(arg: str or float or int) -> bool:
    """
    This function checks if the provided argument can be numeric
    :param arg: the argument that needs to be checked
    :return: a boolean indicating whether the argument is numeric
    """
    try:
        float(arg)
        return True
    except ValueError:
        return False


def check_list_content(input_list: list) -> str:
    """
    This function checks the content of a given list: only numbers, only dictionaries or other.
    :param input_list: the list we need to know the content from
    :return: a string indicating the content of the list
    """
    contains_numbers = all(isinstance(item, (int, float)) for item in input_list)
    contains_dicts = all(isinstance(item, dict) for item in input_list)

    if contains_numbers and not contains_dicts:
        return "numeric"
    if contains_dicts and not contains_numbers:
        return "dictionaries"
    return "other"



# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains all tests for the utils.py file
"""
import pytest
from vlinder.utils import (
    round_all_dict_values,
    get_values_from_target,
    number_formatter,
    check_numeric,
    check_list_content,
)


@pytest.mark.parametrize(
    "my_dict, digits, expected_result",
    [
        (
            {"A": [2.45345, 3.52345], "B": {"C": 3.5356453, "D": [1.243235, 2.4]}},
            3,
            {"A": [2.453, 3.523], "B": {"C": 3.536, "D": [1.243, 2.4]}},
        ),
        (
            {"A": {"B": [2.3243, 3.12], "C": 1.234, "D": {"E": 1.3455}}},
            1,
            {"A": {"B": [2.3, 3.1], "C": 1.2, "D": {"E": 1.3}}},
        ),
    ],
)
def test_round_all_dict_values(my_dict, digits, expected_result):
    """
    This function tests round_all_dict_values to return a rounded dictionary
    :param my_dict: nested input dictionary
    :param digits: number of digits used for rounding
    :param expected_result: expected result of the test
    """
    result = round_all_dict_values(my_dict, digits)
    assert result == expected_result


@pytest.mark.parametrize(
    "dictionary, target, expected_result",
    [
        ({"A": {"B": "Apple", "C": "Banana"}, "B": ["Kiwi", "Mango"]}, "B", ["Apple", ["Kiwi", "Mango"]]),
        ({"A": 1.22, "B": {"A": "Text"}, "C": {"D": {"A": 5}}}, "A", [1.22, "Text", 5]),
    ],
)
def test_get_values_from_target(dictionary, target, expected_result):
    """
    This function tests get_values_from_target to return the right values given the target value
    :param dictionary: nested input dictionary
    :param target: target key of which values need to be collected
    :param expected_result: expected result of the test
    """
    result = get_values_from_target(dictionary, target)
    assert result == expected_result


@pytest.mark.parametrize(
    "number, expected_result",
    [
        (12.456874, "12.46"),
        (755, "755.00"),
        (1005.12, "1.0K"),
        (12567.43, "12.6K"),
        (7684536, "7.7M"),
        (-123456, "-123.5K"),
        (-123.3, "-123.30"),
    ],
)
def test_number_formatter(number, expected_result):
    """
    This function tests number_formatter to return a formatted string
    """
    result = number_formatter(number)
    assert result == expected_result


@pytest.mark.parametrize(
    "arg, expected_result",
    [
        ("256.74", True),
        ("256", True),
        (256, True),
        (256.74, True),
        ("N/A", False),
        ("2/3", False),
        ("23.34.56", False),
    ],
)
def test_check_numeric(arg, expected_result):
    """
    This function tests check_numeric to return the correct boolean
    :param arg: the value to be checked
    :param expected_result: the correct identification of numeric values
    """
    result = check_numeric(arg)
    assert result == expected_result


@pytest.mark.parametrize(
    "arg, expected_result",
    [
        ([23, 1.22, 66], "numeric"),
        ([{"A": 23}, {"B": 23, "C": [1, 2, 3]}], "dictionaries"),
        (["A", {"B": 5}], "other"),
    ],
)
def test_check_list_content(arg, expected_result):
    """
    This function tests check_list_content to correctly identify the content of a list
    :param arg: the list to be checked
    :param expected_result: the expected content of this list
    """
    result = check_list_content(arg)
    assert result == expected_result



# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""This module contains all tests for the Evaluate() class"""
from pathlib import Path
import pytest
from params import INPUT_DICT_BEERWISER
from vlinder.evaluate import Evaluate, EvaluationError
from vlinder.utils import round_all_dict_values
from vlinder.trbs import TheResponsibleBusinessSimulator


@pytest.fixture(name="evaluate_beerwiser")
def fixture_evaluate_beerwiser():
    """
    This fixture initialises a Beerwiser case.
    :return Evaluate(): an Evaluate class for Beerwiser
    """
    return Evaluate(INPUT_DICT_BEERWISER)


@pytest.fixture(name="evaluate_refugee")
def fixture_evaluate_refugee():
    """
    This fixture initialises a Refugee case.
    :return Evaluate(): an Evaluate class for Refugee
    """

    case = TheResponsibleBusinessSimulator("refugee", Path.cwd() / "src/vlinder/data", "xlsx")
    case.build()
    return Evaluate(case.input_dict)


@pytest.fixture(name="evaluate_dsm")
def fixture_evaluate_dsm():
    """
    This fixture initialises a DSM case.
    :return Evaluate(): an Evaluate class for DSM
    """

    case = TheResponsibleBusinessSimulator("DSM", Path.cwd() / "src/vlinder/data")
    case.build()
    return Evaluate(case.input_dict)


@pytest.fixture(name="evaluate_izz")
def fixture_evaluate_izz():
    """
    This fixture initialises a IZZ case.
    :return Evaluate(): an Evaluate class for IZZ
    """

    case = TheResponsibleBusinessSimulator("izz")
    case.build()
    return Evaluate(case.input_dict)


def test_create_value_dict(evaluate_beerwiser):
    """
    This functions tests _create_value_dict returning a correctly initialized value dictionary, for the
    'optimistic' (=1) scenario with 'focus on water recycling' (=2).
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    evaluate_beerwiser._create_value_dict(1, 2)
    result = evaluate_beerwiser.value_dict
    expected_result = {
        "Accidents reduction": 0,
        "Water use reduction": 0,
        "Production cost reduction": 0,
        "Invest in training of employees": 50000,
        "Invest in water recycling": 250000,
        "Cost of accident": 12000.0,
        "Effectiveness water recycling": 1.0,
        "# employees": 500.0,
        "Current # accidents": 51.0,
        "Current production cost": 7500000.0,
        "Current water use": 15000000.0,
        "Water unit cost": 0.05,
    }

    assert result == expected_result


@pytest.mark.parametrize(
    "key, value, expected_result",
    [
        ("key_output_unit", "hl/year", 1),
        ("decision_makers_options", "Equal spread", 0),
        ("operator", "*", 3),
        ("saturation_point", 275000.0, 1),
    ],
)
def test_find_index(evaluate_beerwiser, key, value, expected_result):
    """
    This function tests _find_index to return the FIRST index of a given key and value of input_dict. Tested for a
    variety of cases: different types of 'value', objects where 'value' occurs once or more than once.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    :param key: key in input_dict
    :param value: value in object of input_dict where index needs to be found
    :param expected_result: expected index that is returned by the test
    """
    result = evaluate_beerwiser._find_index(key, value)
    assert result == expected_result


@pytest.mark.parametrize(
    "arg1, args, expected_result",
    [
        (
            10,
            {"saturation_point": 30, "accessibility": 0.9, "probability_of_success": 0.80, "maximum_effect": 0.5},
            0.12,
        ),
        (
            10,
            {"saturation_point": 5, "accessibility": 0.95, "probability_of_success": 0.85, "maximum_effect": 0.7},
            0.565,
        ),
    ],
)
def test_squeeze(evaluate_beerwiser, arg1, args, expected_result):
    """
    This function tests _squeeze to return a correctly calculated values using the Squeezed * operator. Tested for both
    when min(x,y) / saturation_point > 1 and min(x,y) / saturation_point < 1.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    :param arg1: first argument that is used for all operators (including squeezed)
    :param arg2: second argument that is used for all operators (including squeezed)
    :param args: dictionary containing arguments used solely for squeezed
    :param expected_result: expected value of evaluated squeeze function
    """
    result = round(evaluate_beerwiser._squeeze(arg1, args), 3)
    assert result == expected_result


def test_get_key_outputs(evaluate_beerwiser):
    """
    This function tests _get_key_outputs to return a dictionary of type: {KO1: value, KO2: value, ...}.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    # initialise value dictionary | otherwise key_outputs are not in the value dictionary
    evaluate_beerwiser._create_value_dict(1, 2)
    result = evaluate_beerwiser._get_key_outputs()
    expected_result = {"Accidents reduction": 0, "Water use reduction": 0, "Production cost reduction": 0}
    assert result == expected_result


@pytest.mark.parametrize("arg, expected_result", [("2", 2.0), ("3.12", 3.12), ("A", 1.25)])
def test_get_value_of_argument(evaluate_beerwiser, arg, expected_result):
    """
    This function tests _get_value_of_argument to return the right value
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    :param arg: value of argument
    :param expected_result: expected result of this single dependency
    """
    evaluate_beerwiser.value_dict = {"A": 1.25}
    result = evaluate_beerwiser._get_value_of_argument(arg)
    assert result == expected_result


@pytest.mark.parametrize(
    "arg1, arg2, operator, expected_result",
    [
        (8, 12, "-", -4),
        (22, 19, "+", 41),
        (13, 7, "*", 91),
        (20, 0, "/", 0),
        (99, 9, "/", 11),
        (12, 8, "-*", -96),
        (121, -11, "-/", 11),
        (0, 0, "-/", 0),
        (5, 10, "<", 1),
        (15, 10, ">", 1),
        (10, 10, "<=", 1),
        (5, 10, ">=", 0),
        (2, 3, "min", 2),
        (10, 10.5, "max", 10.5),
    ],
)
def test_evaluate_single_dependency(evaluate_beerwiser, arg1, arg2, operator, expected_result):
    """
    This function tests _evaluate_single_dependency to return the proper value for ALL allowed operators. Also, the
    exemption case for division (division by zero should return zero) is tested.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    :param arg1: value of first argument
    :param arg2: value of second argument
    :param operator: operator needed to calculate result based on the two arguments
    :param expected_result: expected result of this single dependency
    """
    result = evaluate_beerwiser._evaluate_single_dependency(arg1, arg2, operator)
    assert result == expected_result


def test_evaluate_single_dependency_evaluation_error(evaluate_beerwiser):
    """
    This function tests _evaluate_single_dependency to raise an EvaluationError when an undefined operator is used.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    with pytest.raises(EvaluationError) as evaluation_error:
        evaluate_beerwiser._evaluate_single_dependency(3, 8, "/*")
    expected_result = "Evaluation Error: operator /* not available"
    assert str(evaluation_error.value) == expected_result


@pytest.mark.parametrize(
    "fixture_name, scenario, dmo, expected_result",
    [
        (
            "evaluate_beerwiser",
            "Optimistic",
            "Focus on water recycling",
            {
                "key_outputs": {
                    "Accidents reduction": 3.49,
                    "Water use reduction": 6818181.82,
                    "Production cost reduction": 0.05,
                }
            },
        ),
        (
            "evaluate_refugee",
            "Base case",
            "Employment & Language",
            {
                "key_outputs": {
                    "Quality of life Ukrainians": 0.07,
                    "Quality of life other refugees": -0.03,
                    "Quality of life inhabitants": -0.05,
                    "Unemployment reduction Ukrainians": 0.21,
                    "Unemployment reduction other refugees": 0.03,
                    "Unemployment reduction inhabitants": 0.01,
                    "Economic impact": 23857682.88,
                }
            },
        ),
        (
            "evaluate_dsm",
            "Base case",
            "Partner RE",
            {
                "key_outputs": {
                    "CAPEX": 0.0,
                    "Energy cost": 4335000.0,
                    "RE %": 0.5,
                    "Carbon footprint reduction": 10140.0,
                    "Actual carbon emission": 10140.0,
                    "Increase in employee engagement score": 0.02,
                    "Increase in employee recommendation rate": 0.01,
                    "Increase in Net Promotor Score": 0,
                    "Increase in brand value": 2149680.0,
                }
            },
        ),
        (
            "evaluate_izz",
            "Optimistic",
            "Social",
            {
                "key_outputs": {
                    "Decrease in absenteeism %": 0.0,
                    "Decrease in staff turnover %": 0.0,
                    "Increase in customer satisfaction": 0.01,
                    "Increase in employee satisfaction": 0.03,
                    "Decrease in absenteeism costs": 18610.66,
                    "Decrease in staff turnover costs": 734.7,
                    "Decrease in wage costs": 11235861.75,
                    "Increase in production capacity": 513.44,
                    "Total investment": 120000.0,
                }
            },
        ),
    ],
)
def test_evaluate_all_dependencies(fixture_name, expected_result, scenario, dmo, request):
    """
    This function tests evaluate_all_dependencies to return a dictionary containing the key_outputs for a given
    scenario and decision_makers_option.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    case = request.getfixturevalue(fixture_name)
    result = case.evaluate_all_dependencies(scenario, dmo)
    rounded_result = round_all_dict_values(result)
    assert rounded_result == expected_result


def test_evaluate_selected_scenario_only_structure(evaluate_beerwiser):
    """
    This function tests the STRUCTURE of the dictionary returned by evaluate_selected_scenario(). Only the keys and the
    structure of the values are checked as these values are a result from evaluate_all_dependencies() which is already
    checked above.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    result = evaluate_beerwiser.evaluate_selected_scenario("Optimistic")
    result_structure = {key: type(value) for key, value in result.items()}
    expected_structure = {"Equal spread": dict, "Focus on training": dict, "Focus on water recycling": dict}
    assert result_structure == expected_structure


def test_evaluate_all_scenarios_only_structure(evaluate_beerwiser):
    """
    This function tests the STRUCTURE of the dictionary returned by evaluate_all_scenarios(). Only the keys and the
    structure of the values are checked as these values are a result from evaluate_selected_scenario() which is already
    checked above.
    :param evaluate_beerwiser: an Evaluate() class for Beerwiser
    """
    result = evaluate_beerwiser.evaluate_all_scenarios()
    result_structure = {key: type(value) for key, value in result.items()}
    expected_structure = {"Base case": dict, "Optimistic": dict, "Pessimistic": dict}
    assert result_structure == expected_structure



"""
This file contains the Evaluate class that deals with the calculation of key output values for
the decision makers options.
"""
import numpy as np


class EvaluationError(Exception):
    """
    This class deals with the error handling of our evaluate calculations.
    """

    def __init__(self, message):  # ignore warning about super-init | pylint: disable=W0231
        self.message = message

    def __str__(self):
        return f"Evaluation Error: {self.message}"


class Evaluate:
    """This class deals with the calculation of key output values for the decision makers options."""

    def __init__(self, input_dict):
        self.input_dict = input_dict
        self.value_dict = {}
        # dictionary containing all functions to which operators are related
        self.operators_dict = {
            "-": lambda x, y: x - y,
            "+": lambda x, y: x + y,
            "*": lambda x, y: x * y,
            "/": lambda x, y: x / y if y else 0,
            "-*": lambda x, y: -x * y,
            "-/": lambda x, y: -x / y if y else 0,
            ">": lambda x, y: 1 if x > y else 0,
            "<": lambda x, y: 1 if x < y else 0,
            ">=": lambda x, y: 1 if x >= y else 0,
            "<=": lambda x, y: 1 if x <= y else 0,
            "min": lambda x, y: min(x, y),  # ignore warning about unnecessary lambda | pylint: disable=W0108
            "max": lambda x, y: max(x, y),  # ignore warning about unnecessary lambda | pylint: disable=W0108
        }

    def _create_value_dict(self, scen_index: int, dmo_index: int) -> None:
        """
        This function creates a 'value-dictionary' that maps all variables to their current value.
        :param scen_index: index of the current scenario in the input_dictionary
        :param dmo_index: index of the current decision makers option in the input_dictionary
        :return None: self.value_dict is updated in this function
        """
        self.value_dict = {
            # add key outputs. Initialise at zero.
            **{key: 0 for key in self.input_dict["key_outputs"]},
            # add internal variable values (for given dmo and scenario)
            **dict(
                zip(
                    self.input_dict["internal_variable_inputs"],
                    self.input_dict["decision_makers_option_value"][dmo_index],
                )
            ),
            # add external variable values
            **dict(zip(self.input_dict["external_variable_inputs"], self.input_dict["scenario_value"][scen_index])),
            # add fixed values
            **dict(zip(self.input_dict["fixed_inputs"], self.input_dict["fixed_input_value"])),
        }

    def _find_index(self, key: str, value: str or int) -> int:
        """This helper function returns the FIRST index of a value for a given key and value of self.input_dict."""
        return np.where(self.input_dict[key] == value)[0][0]

    def _squeeze(self, argument_1_value: float, squeeze_args: dict) -> int:
        """This functions evaluates ONLY the squeeze * operator function."""
        division_part = self.operators_dict["/"](argument_1_value, squeeze_args["saturation_point"])
        result = (
            min(1, division_part)
            * squeeze_args["accessibility"]
            * squeeze_args["probability_of_success"]
            * squeeze_args["maximum_effect"]
        )
        return result

    def _get_key_outputs(self) -> dict:
        """
        This function retrieves the current values of the key outputs, based on the value_dict.
        :return: a dictionary matching key outputs to their values, i.e. {KO1: 1.12, KO2: 2.33, ...}
        """
        key_output_values = [self.value_dict[key_output] for key_output in self.input_dict["key_outputs"]]
        return dict(zip(self.input_dict["key_outputs"], key_output_values))

    def _get_value_of_argument(self, arg: str) -> float:
        """
        This function transforms a string to a number (if applicable) else it looks up the correct value in the
        value dictionary.
        :param arg: the value that needs to be checked
        :return: the correct value of the argument
        """
        try:
            value = float(arg)
        except ValueError:
            value = self.value_dict[arg]
        return value

    def _evaluate_single_dependency(self, argument_1_value: float, argument_2_value: float, operator: str) -> int:
        """
        This function evaluates a single dependency. Raises an EvaluationError when the operator is unknown.
        :param argument_1_value: value of first argument
        :param argument_2_value: value of second argument
        :param operator: operator needed to calculate result based on the two arguments
        :return result: calculated value based on inputs & operator
        """
        if operator not in self.operators_dict.keys():  # ignore warning about .keys() | pylint: disable=C0201
            raise EvaluationError(f"operator {operator} not available")

        # apply operations based on occurrence in operators_dict
        result = self.operators_dict[operator](argument_1_value, argument_2_value)
        return result

    def evaluate_all_dependencies(self, scenario: str, decision_makers_option: str) -> dict:
        """
        This function returns an output dictionary containing the values of the key outputs for a given scenario and
        a given decision makers option.
        :param scenario: string of scenario name
        :param decision_makers_option: string of decision makers option
        :return: dictionary containing all key outputs for given scenario and decision makers option
        """
        scen_index = self._find_index("scenarios", scenario)
        dmo_index = self._find_index("decision_makers_options", decision_makers_option)
        self._create_value_dict(scen_index, dmo_index)

        # calculate each destination -- already ordered on hierarchy during the import
        for index, dest in enumerate(self.input_dict["destination"]):
            dest_value = self.value_dict.get(dest, 0)
            arg1 = self.input_dict["argument_1"][index]
            arg2 = self.input_dict["argument_2"][index]
            operator = self.input_dict["operator"][index]

            # squeezed * has its own evaluation function
            if operator == "squeezed *":
                dest_result = dest_value + self._squeeze(
                    self._get_value_of_argument(arg1),
                    {
                        "saturation_point": self.input_dict["saturation_point"][index],
                        "accessibility": self.input_dict["accessibility"][index],
                        "probability_of_success": self.input_dict["probability_of_success"][index],
                        "maximum_effect": self.input_dict["maximum_effect"][index],
                    },
                )
            # All other operators are calculated using a general approach
            else:
                dest_result = dest_value + self._evaluate_single_dependency(
                    self._get_value_of_argument(arg1), self._get_value_of_argument(arg2), operator
                )

            # update the value dictionary
            result = {dest: dest_result}
            self.value_dict.update(result)

        # structure the output dictionary
        output_dict = {"key_outputs": self._get_key_outputs()}
        return output_dict

    def evaluate_selected_scenario(self, scenario: str) -> dict:
        """
        This function creates an output dictionary for all decision makers option within a given scenario.
        :param scenario:
        :return: a dictionary that contains the outputs per decision makers options (for a single scenario)
        """
        output_dict = {}
        for decision_makers_option in self.input_dict["decision_makers_options"]:
            output_dict[decision_makers_option] = self.evaluate_all_dependencies(scenario, decision_makers_option)

        return output_dict

    def evaluate_all_scenarios(self) -> dict:
        """
        This function evaluates the dependencies for all scenario's and all decision makers options.
        :return: for each scenario, a dictionary for all decision makers options is returned
        """
        output_dict = {}
        for scenario in self.input_dict["scenarios"]:
            output_dict[scenario] = self.evaluate_selected_scenario(scenario)
            print(f"- Evaluated '{scenario}' successfully for all decision makers options!")

        return output_dict



"""
This file contains the Appreciate class that deals with the calculation of appreciations.
"""
import math
import pandas as pd
import numpy as np
from vlinder.utils import get_values_from_target


class Appreciate:
    """This class deals with the calculation of appreciations"""

    def __init__(self, input_dict, output_dict):
        self.input_dict = input_dict
        self.output_dict = output_dict
        self.start_and_end_points = self._get_start_and_end_points()

    # TODO: monetary values
    def _get_start_and_end_points(self) -> dict:
        """
        This function obtains the minimum and maximum values of the calculated key_output values, over ALL scenarios
        and decision makers options.
        :return: a dictionary of type {key_output 1: [min, max], key_output 2: [min, max], .. }
        """
        boundaries = {}
        all_key_output_values = get_values_from_target(self.output_dict, "key_outputs")
        values_as_df = pd.DataFrame.from_dict(all_key_output_values)

        for key_output in values_as_df.columns:
            boundaries[key_output] = [values_as_df[key_output].min(), values_as_df[key_output].max()]

        return boundaries

    def _appreciate_single_key_output(self, value: float, args: dict) -> float:
        """
        This function appreciates a single key output, given the value, arguments and start & end point.
        :param value: value of key output
        :param args: dictionary containing whether the key output should be appreciated linear or smaller the better.
        :return: the appreciated value of the key output
        """
        start_and_end = self.start_and_end_points[args["key_output"]]
        stb_ind = args["key_output_smaller_the_better"]

        # Option 0A: values lie outside the boundaries. Return maximum or minimum based on STB
        if value < start_and_end[0] or value > start_and_end[1]:
            return stb_ind * (value < start_and_end[0]) * 100 + (1 - stb_ind) * (value > start_and_end[1]) * 100

        # Option 0B: start and end value are the same --> indifferent so return 0
        if start_and_end[1] - start_and_end[0] < 1e-6:
            return 0

        # Option 1: Linear appreciation
        #   - if STB = 1: (end - val) / (end - start) * 100
        #   - if STB = 0: (val - start) / (end - start) * 100
        if args["key_output_linear"]:
            return [-1, 1][stb_ind] * (start_and_end[stb_ind] - value) / (start_and_end[1] - start_and_end[0]) * 100
        # Option 2: Non-linear appreciation
        #   - if STB = 1: (-sin(0.5 * pi * (val-start) / (end - start) + 1) * 100
        #   - if STB = 0: sin(0.5 * pi * (val - start) / (end - start) + 0) * 100
        core_part = (value - start_and_end[0]) / (start_and_end[1] - start_and_end[0])
        return ([1, -1][stb_ind] * math.sin(0.5 * math.pi * core_part) + stb_ind) * 100

    def appreciate_single_decision_maker_option(self, value_dict_in: dict) -> None:
        """
        This function calculates the appreciation values, both weighted as well as unweighted for the key outputs for a
        given scenario and decision makers option. Results are stored within the output_dict.
        :param scenario: given scenario
        :param decision_maker_option: given dmo
        :param value_dict_in: dictionary corresponding with given scenario and dmo
        :return: None as results are stored within the output_dict
        """
        value_dict_in["appreciations"] = {}
        for index, key_output in enumerate(self.input_dict["key_outputs"]):
            key_output_value = value_dict_in["key_outputs"][key_output]
            appreciation_args = {
                "key_output": key_output,
                "key_output_smaller_the_better": self.input_dict["key_output_smaller_the_better"][index],
                "key_output_linear": self.input_dict["key_output_linear"][index],
            }
            appreciation = self._appreciate_single_key_output(key_output_value, appreciation_args)

            value_dict_in["appreciations"][key_output] = appreciation

        weighted_appreciations = self._apply_weights(value_dict_in["appreciations"])
        value_dict_in["weighted_appreciations"] = {
            key: weighted_appreciations[index] for index, key in enumerate(self.input_dict["key_outputs"])
        }
        value_dict_in["decision_makers_option_appreciation"] = sum(weighted_appreciations)

    def appreciate_single_scenario(self, value_dict_in: dict) -> None:
        """
        This function calculates the appreciation values, both weighted as well as unweighted for the key outputs for a
        given scenario and ALL decision makers options. Results are stored within the output_dict.
        :param scenario: given scenario
        :param value_dict_in: dictionary corresponding with given scenario
        :return: None as results are stored within the output_dict
        """
        for _, value_dict_out in value_dict_in.items():
            self.appreciate_single_decision_maker_option(value_dict_out)

    def appreciate_all_scenarios(self) -> None:
        """
        This function calculates the appreciation values, both weighted as well as unweighted for the key outputs for a
        given scenario and ALL decision makers options. Results are stored within the output_dict.
        :return: None as results are stored within the output_dict
        """ ""
        for _, value_dict in self.output_dict.items():
            self.appreciate_single_scenario(value_dict)
        self._apply_scenario_weights()
        print("Key output values have been processed | Appreciated, weighted & aggregated")

    @staticmethod
    def _apply_weights_single_key_output(weights: dict) -> float:
        """
        This function calculates the weight for a single key output.
        :param weights: dictionary containing necessary parameters. Should contain:
            - 'sum_within_theme': sum of all weights of all key outputs with the same theme
            - 'sum_theme': sum of the theme weights (theme weights are user input)
            - 'key_output': weight of this key_output
            - 'theme': weight of the theme of the key_output
        :return: the calculated weight of the key output
        """
        # 'sum_within_theme' or 'sum_theme' cannot be 0
        if not weights["sum_within_theme"] or not weights["sum_theme"]:
            return 0
        return (weights["key_output"] / weights["sum_within_theme"]) * (weights["theme"] / weights["sum_theme"])

    def _calculate_weights(self) -> list:
        """
        This function creates a weights list for all key outputs.
        :return: list with weights for all key outputs
        """
        key_output_themes = self.input_dict["key_output_theme"]
        weights = self.input_dict["key_output_weight"]
        theme_weights = self.input_dict["theme_weight"]

        adjusted_weights = []
        for index, _ in enumerate(self.input_dict["key_outputs"]):
            theme = key_output_themes[index]
            # Find all key outputs weights that share the same theme as 'key_output' (>= 1)
            theme_indices = [
                theme_index for theme_index, theme_value in enumerate(key_output_themes) if theme_value == theme
            ]
            weights_within_theme = weights[theme_indices]
            # Find theme weight of 'key_output' - as a number instead of array
            theme_weight = theme_weights[np.where(self.input_dict["themes"] == theme)[0]][0]

            weights_dict = {
                "key_output": weights[index],
                "theme": theme_weight,
                "sum_theme": np.sum(theme_weights),
                "sum_within_theme": np.sum(weights_within_theme),
            }
            adjusted_weights.append(self._apply_weights_single_key_output(weights_dict))

        return adjusted_weights

    def _apply_weights(self, appreciation_dict: dict) -> np.array:
        """
        This function applies the weights to the appreciations
        :param appreciation_dict: dictionary contain the unweighted appreciations
        :return: a numpy array with the weighted appreciations
        """
        # ensure appreciation vector has same order as weight & key_outputs vector
        appreciations = [appreciation_dict[key_output] for key_output in self.input_dict["key_outputs"]]
        # calculate the (adjusted) weight for given key_output & theme weight
        weights = self._calculate_weights()
        weighted_appreciations = np.array(appreciations) * np.array(weights)

        return weighted_appreciations

    def _apply_scenario_weights(self) -> None:
        """
        This function applies scenario weights to the decision makers' option appreciations
        :return: None as results are stored within the output_dict
        """
        total_weight = sum(self.input_dict["scenario_weight"])

        for scenario, weight in zip(self.input_dict["scenarios"], self.input_dict["scenario_weight"]):
            for option in self.input_dict["decision_makers_options"]:
                appreciation = self.output_dict[scenario][option]["decision_makers_option_appreciation"]
                weighted_appreciation = appreciation * weight / total_weight
                self.output_dict[scenario][option]["scenario_appreciations"] = weighted_appreciation



# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains all tests for the CaseImporter() class.
NOTE: _create_input_dict & import_case are not tested as they are only helper function that call other functions
"""
import warnings
from pathlib import Path
import pytest
import pandas as pd
import numpy as np
from vlinder.case_importer import CaseImporter, TemplateError


@pytest.fixture(name="import_beerwiser_json")
def fixture_import_beerwiser_json():
    """
    This fixture initialises a CaseImporter for the beerwiser case.
    :return: an CaseImporter class for the beerwiser case.
    """
    return CaseImporter(Path.cwd() / "src/vlinder/data", "beerwiser", "json")


@pytest.fixture(name="import_beerwiser_csv")
def fixture_import_beerwiser_csv():
    """
    This fixture initialises a CaseImporter for the beerwiser case.
    :return: an CaseImporter class for the beerwiser case.
    """
    return CaseImporter(Path.cwd() / "src/vlinder/data", "beerwiser", "csv")


@pytest.fixture(name="import_beerwiser_xlsx")
def fixture_import_beerwiser_xlsx():
    """
    This fixture initialises a CaseImporter for the beerwiser case.
    :return: an CaseImporter class for the beerwiser case.
    """
    return CaseImporter(Path.cwd() / "src/vlinder/data", "beerwiser", "xlsx")


def test_build_template_validators(import_beerwiser_json):
    """
    This function tests _build_template_validators() to return a dictionary containing the 'table' and necessary
    columns within that table.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    result = import_beerwiser_json._build_template_validators()
    expected_result = {
        "configurations": ["configuration", "value"],
        "generic_text_elements": ["generic_text_element", "value"],
        "case_text_elements": ["case_text_element", "value"],
        "key_outputs": [
            "key_output",
            "theme",
            "monetary",
            "smaller_the_better",
            "linear",
            "automatic",
            "start",
            "end",
            "threshold",
        ],
        "decision_makers_options": ["internal_variable_input", "decision_makers_option", "value"],
        "scenarios": ["external_variable_input", "scenario", "value"],
        "fixed_inputs": ["fixed_input", "value"],
        "dependencies": [
            "destination",
            "argument_1",
            "argument_2",
            "operator",
            "maximum_effect",
            "accessibility",
            "probability_of_success",
            "saturation_point",
        ],
        "theme_weights": ["theme", "weight"],
        "key_output_weights": ["key_output", "weight"],
        "scenario_weights": ["scenario", "weight"],
    }
    assert result == expected_result


@pytest.mark.parametrize(
    "dataframe, table, expected_result",
    [
        (
            pd.DataFrame(columns=["theme", "weight", "EXTRA COL"]),
            "theme_weights",
            pd.DataFrame(columns=["theme", "weight"]),
        ),
        (
            pd.DataFrame(columns=["external_variable_input", "scenario", "value"]),
            "scenarios",
            pd.DataFrame(columns=["external_variable_input", "scenario", "value"]),
        ),
    ],
)
def test_check_data_columns(import_beerwiser_json, dataframe, table, expected_result):
    """
    This function tests _check_data_columns to return the same pd.DataFrame with ONLY the necessary columns. Warnings
    are tested separately below.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    :param dataframe: dataframe where the columns need to be checked
    :param table: table name of the dataframe
    :expected_result: dataframe that should be returned by _check_data_columns
    """
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        result = import_beerwiser_json._check_data_columns(dataframe, table)

    pd.testing.assert_frame_equal(result, expected_result)


def test_check_data_columns_warning(import_beerwiser_json):
    """
    This function tests _check_data_columns to raise a warning when extra columns are provided.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    with warnings.catch_warnings(record=True) as warning_list:
        import_beerwiser_json._check_data_columns(
            pd.DataFrame(columns=["fixed_input", "value", "EXTRA COL"]), "fixed_inputs"
        )
    expected_result = "column(s) 'EXTRA COL' are not used for 'fixed_inputs'"
    assert str(warning_list[0].message) == expected_result


def test_check_data_columns_error(import_beerwiser_json):
    """
    This function tests _check_data_column to raise a TemplateError when necessary columns are missing
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    with pytest.raises(TemplateError) as template_error:
        import_beerwiser_json._check_data_columns(
            pd.DataFrame(columns=["internal_variable_input", "value"]), "decision_makers_options"
        )
    expected_result = "Template Error: column(s) 'decision_makers_option' are missing for 'decision_makers_options'"
    assert str(template_error.value) == expected_result


@pytest.mark.parametrize(
    "fixture_name, table",
    [
        ("import_beerwiser_csv", "dependencies"),
        ("import_beerwiser_json", "key_outputs"),
        ("import_beerwiser_xlsx", "scenarios"),
    ],
)
def test_create_dataframes_dict(fixture_name, table, request):
    """
    This function tests _create_dataframes to update self.dataframes_dict with a new {key: pd.DataFrame} pair.
    Only the structure of the output is tested.
    :param fixture_name: used version of the CaseImporter class. Fixtures differ in used extension / file format
    :param table: name of the table
    :param request: needed to transform fixture_name to a fixture value
    """
    case_importer = request.getfixturevalue(fixture_name)
    case_importer._create_dataframes_dict(table)
    result_structure = {key: type(value) for key, value in case_importer.dataframes_dict.items()}
    expected_structure = {table: pd.DataFrame}
    assert result_structure == expected_structure


def test_create_dataframes_dict_error(import_beerwiser_json):
    """
    This function tests _create_dataframes_dict to raise an error when a missing table name is called.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    with pytest.raises(TemplateError) as template_error:
        import_beerwiser_json._create_dataframes_dict("Missing Table")
    expected_result = "Template Error: Sheet 'Missing Table' is missing"
    assert str(template_error.value) == expected_result


def test_convert_to_numpy_arrays_2d(import_beerwiser_json):
    """
    This function tests _convert_to_numpy_arrays_2d to return properly pivoted numpy arrays from the given dataframe.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    input_dataframe = pd.DataFrame(
        {
            "internal_variable_input": ["A", "B", "A", "B"],
            "decision_makers_option": ["X", "X", "Y", "Y"],
            "value": [2, 3, 4, 5],
        }
    )
    import_beerwiser_json._convert_to_numpy_arrays_2d("decision_makers_options", input_dataframe)
    result = import_beerwiser_json.input_dict
    expected_result = {
        "decision_makers_options": np.array(["X", "Y"]),
        "internal_variable_inputs": np.array(["A", "B"]),
        "decision_makers_option_value": np.array([[2, 3], [4, 5]]),
    }
    assert all(np.array_equal(expected_result[key], result[key]) for key in result)


@pytest.mark.parametrize(
    "row, all_inputs, expected_output",
    [
        (pd.Series({"argument_1": "money", "argument_2": "love"}), np.array(["education", "money"]), 2),
        (pd.Series({"argument_1": "love", "argument_2": "love"}), np.array(["education", "police", "law"]), 2),
        (pd.Series({"argument_1": "law", "argument_2": "love"}), np.array(["love", "money", "law"]), 1),
    ],
)
def test_apply_first_level_hierarchy_to_row(import_beerwiser_json, row, all_inputs, expected_output):
    """
    This function tests _apply_first_level_hierarchy_to_row to return a 1 or 2 based on the presence of names in the
    input array. Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    :param row: a pd.Series containing the needed values from dependency row
    :param all_inputs: an array containing the available input names
    :param expected_output: the expected output of the function
    """
    result = import_beerwiser_json._apply_first_level_hierarchy_to_row(row, all_inputs)
    assert result == expected_output


@pytest.mark.parametrize(
    "row, expected_output",
    [
        (pd.Series({"destination": "law", "argument_1": "money", "argument_2": "love", "hierarchy": 1}), 1),
        (pd.Series({"destination": "education", "argument_1": "money", "argument_2": "love", "hierarchy": 3}), 5),
    ],
)
def test_apply_second_level_hierarchy_to_row(import_beerwiser_json, row, expected_output):
    """
    This function tests _apply_second_level_hierarchy_to_row to return the right hierarchy level. For testing purposes
    only the relevant columns are provided as input.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    :param row: single row of dependencies
    :param data: a dataframe containing all dependencies
    """
    input_dataframe = pd.DataFrame({"destination": np.array(["love", "law", "law", "money", "education"])})
    result = import_beerwiser_json._apply_second_level_hierarchy_to_row(row, input_dataframe)
    assert result == expected_output


def test_convert_to_ordered_dependencies(import_beerwiser_json):
    """
    This function tests test_convert_to_ordered_dependencies to add the dependencies in the correct order to the
    input dictionary
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    # set-up for testing
    import_beerwiser_json.input_dict = {
        "fixed_inputs": ["love", "money"],
        "internal_variable_inputs": [],
        "external_variable_inputs": [],
    }
    input_data = pd.DataFrame(
        {
            "destination": ["law", "super love", "lawyer", "IT", "crook", "education"],
            "argument_1": ["tech", "love", "education", "law", "5", "money"],
            "argument_2": ["", "", "law", "tech", "money", "love"],
            "operator": np.full(6, "N/A"),
            "maximum_effect": np.full(6, "N/A"),
            "accessibility": np.full(6, "N/A"),
            "probability_of_success": np.full(6, "N/A"),
            "saturation_point": np.full(6, "N/A"),
        }
    )
    # compare
    import_beerwiser_json._convert_to_ordered_dependencies(input_data)
    result = import_beerwiser_json.input_dict
    expected_result = {
        "fixed_inputs": ["love", "money"],
        "internal_variable_inputs": [],
        "external_variable_inputs": [],
        "destination": np.array(["crook", "education", "law", "super love", "lawyer", "IT"], dtype=object),
        "argument_1": np.array(["5", "money", "tech", "love", "education", "law"], dtype=object),
        "argument_2": np.array(["money", "love", "", "", "law", "tech"], dtype=object),
        "operator": np.full(6, "N/A", dtype=object),
        "maximum_effect": np.full(6, "N/A", dtype=object),
        "accessibility": np.full(6, "N/A", dtype=object),
        "probability_of_success": np.full(6, "N/A", dtype=object),
        "saturation_point": np.full(6, "N/A", dtype=object),
        "hierarchy": np.array([1, 1, 2, 2, 3, 3]),
        "dependencies_order": np.array([4, 5, 0, 1, 2, 3]),
    }
    assert all(np.array_equal(expected_result[key], result[key]) for key in result)


def test_convert_to_numpy_arrays_weights(import_beerwiser_json):
    """
    This function tests _convert_to_numpy_arrays_weights to properly update the input dictionary.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    import_beerwiser_json.input_dict = {"key_outputs": ["C", "B", "A"]}
    input_data = pd.DataFrame({"key_output": ["A", "B", "C"], "weight": [1, 5, 2]})
    import_beerwiser_json._convert_to_numpy_arrays_weights("key_output_weights", input_data)
    result = import_beerwiser_json.input_dict
    expected_result = {"key_outputs": np.array(["C", "B", "A"]), "key_output_weight": np.array([2, 5, 1])}
    assert all(np.array_equal(expected_result[key], result[key]) for key in result)


def test_convert_to_numpy_arrays(import_beerwiser_json):
    """
    This function tests _convert_to_numpy_arrays to properly update the input dictionary
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    input_dataframe = pd.DataFrame({"fixed_input": ["A", "B", "C", "D"], "value": [3, 5, 2, 8]})
    import_beerwiser_json._convert_to_numpy_arrays("fixed_inputs", input_dataframe)
    result = import_beerwiser_json.input_dict
    expected_result = {"fixed_inputs": np.array(["A", "B", "C", "D"]), "fixed_input_value": np.array([3, 5, 2, 8])}
    assert all(np.array_equal(expected_result[key], result[key]) for key in result)


@pytest.mark.parametrize(
    "input_dataframe, expected_output",
    [
        (
            {
                "key_output_theme": ["People", "Planet", "Profit"],
                "key_output_weight": [2, 1, 3],
                "themes": ["Planet", "People", "Profit"],
                "theme_weight": [1, 2, 3],
            },
            {
                "key_output_theme": ["People", "Planet", "Profit"],
                "key_output_weight": [2, 1, 3],
                "themes": ["Planet", "People", "Profit"],
                "theme_weight": [1, 2, 3],
                "key_output_relative_weight": np.array([2, 1, 3]),
            },
        ),
        (
            {
                "key_output_theme": ["Society", "Society", "Society", "Society", "Society", "Society", "Economy"],
                "key_output_weight": [1, 1, 1, 1, 1, 1, 1],
                "themes": ["Economy", "Society"],
                "theme_weight": [1, 1],
            },
            {
                "key_output_theme": ["Society", "Society", "Society", "Society", "Society", "Society", "Economy"],
                "key_output_weight": [1, 1, 1, 1, 1, 1, 1],
                "themes": ["Economy", "Society"],
                "theme_weight": [1, 1],
                "key_output_relative_weight": np.array(
                    [0.16666667, 0.16666667, 0.16666667, 0.16666667, 0.16666667, 0.16666667, 1]
                ),
            },
        ),
    ],
)
def test_convert_to_relative_weights(import_beerwiser_json, input_dataframe, expected_output):
    """
    This function tests _convert_to_relative_weights to properly update the input dictionary.
    Note: the file extension (here: json) is not relevant for this test, so only one needs to be tested
    :param import_beerwiser_json: an CaseImporter class for the beerwiser case
    """
    import_beerwiser_json.input_dict = input_dataframe
    import_beerwiser_json._convert_to_relative_weights()
    result = import_beerwiser_json.input_dict
    assert np.allclose(result["key_output_relative_weight"], expected_output["key_output_relative_weight"])
    for key in ["key_output_theme", "key_output_weight", "themes", "theme_weight"]:
        assert result[key] == expected_output[key]



"""
This file contains the Visualize class that deals with the creation of all graphs and tables
"""
import re
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from vlinder.utils import round_all_dict_values, number_formatter, get_values_from_target, check_list_content


class VisualizationError(Exception):
    """
    This class deals with the error handling of the Visualize class
    """

    def __init__(self, message):  # ignore warning about super-init | pylint: disable=W0231
        self.message = message

    def __str__(self):
        return f"Visualization Error: {self.message}"


class Visualize:
    """This class deals with the creation of all graphs and tables"""

    def __init__(self, outcomes, options):
        # for visualization purposes two digits is sufficient
        self.outcomes = round_all_dict_values(outcomes)
        self.options = options
        self.colors = [
            "#295477",
            "#F3DD8C",
            "#D04A02",
            "#299D8F",
            "#E72B33",
            "#6A91B6",
            "#EB8C00",
            "#FFB600",
            "#5CB9BD",
            "#E27588",
        ]
        self.available_visuals = {
            "table": self._create_table,
            "barchart": self._create_barchart,
        }
        self.available_outputs = [
            "key_outputs",
            "appreciations",
            "weighted_appreciations",
            "decision_makers_option_appreciation",
        ]
        self.available_kwargs = ["scenario", "decision_makers_option", "stacked", "show_legend"]

    def _validate_kwargs(self, **kwargs) -> None:
        """
        This function validates whether the provided additional arguments are relevant.
        :param **kwargs: the additional arguments provided
        :return: None if all **kwargs are available, else raise an VisualizationError.
        """
        for argument, _ in kwargs.items():
            if argument not in self.available_kwargs:
                raise VisualizationError(f"Invalid argument '{argument}'")

    def _find_dimension_level(self, my_dict: dict, target_key: str, level: int = 1) -> int or None:
        """
        This recursive function returns the dimension level (level of nesting) for a given dictionary and target key.
        For example in dictionary {A: {B: {C: 1.23, ..}, ..}, ..}. 'A' is nested at level 1, B at level 2 & C level 3.
        :param my_dict: input dictionary
        :param target_key: name of the key
        :param level: current dimension level
        :return: the dimension level (if target_key exists)
        """
        if target_key in my_dict:
            return level

        for _, value in my_dict.items():
            if isinstance(value, dict):
                found_level = self._find_dimension_level(value, target_key, level + 1)
                if found_level is not None:
                    return found_level
        return None

    @staticmethod
    def _str_snake_case_to_text(snake_case_str: str) -> str:
        """
        This styling function returns a formatted string where the snake case is replaced by spaces.
        :param snake_case_str: string to reformat
        :result: string where _ is replaced by spaces
        """
        words = snake_case_str.split("_")
        return " ".join(words)

    @staticmethod
    def _truncate_title_list(title_list: list) -> list:
        """
        This function truncates a list of strings to ensure that in total they have no more than 85 chars. For example,
        two strings could have 42 chars whereas, 10 strings can only use 8 chars.
        :param title_list: list of title names to be truncated
        :return: a truncated (where necessary) list of titles
        """
        max_char_length = int(85 / len(title_list))
        truncated_list = [
            f"{item[:(max_char_length-2)]}.." if len(item) > max_char_length else item for item in title_list
        ]
        return truncated_list

    @staticmethod
    def _table_styler(styler: pd.DataFrame.style, table_name: str) -> pd.DataFrame.style:
        """
        This function adds a coherent style for all generated tables
        :param styler: a Pandas styler object
        :param table_name: name of the styled table
        :return: an updated styler object
        """
        # set color palette
        cmap = mpl.cm.Blues(np.linspace(0, 1, 30))
        cmap = mpl.colors.ListedColormap(cmap[:8, :-1])

        # style
        styler.format(number_formatter)
        styler.set_caption(table_name)
        styler.background_gradient(cmap=cmap, axis=1)
        return styler

    def _graph_styler(self, axis: mpl.axis, title: str, show_legend: bool) -> mpl.axis:
        """
        This function adds a coherent style for all generated graphs.
        :param axis: a matplotlib axis object containing the graph
        :param title: title of the graph
        :return: formatted matplotlib axis object
        """
        xtick_formatted = self._truncate_title_list([label.get_text() for label in axis.get_xticklabels()])
        axis.set_title(title, color="#777777", fontsize=12)
        axis.set_ylim(0, 100)
        axis.set_xticklabels(xtick_formatted, rotation=0, fontweight="bold")
        axis.legend(loc="upper left", bbox_to_anchor=(1, 1))
        if not show_legend:
            axis.legend_ = None
        return axis

    def _format_data_for_visual(self, key_data: str) -> pd.DataFrame:
        """
        This function formats the values of the given key into a dataframe.
        :param key_data: name of the key that needs formatting
        :return: a pd.DataFrame with decision maker options as row index, key outputs as columns and weighted
        appreciations as values.
        """
        dim_level = self._find_dimension_level(self.outcomes, key_data)
        dim_names = ["scenario", "decision_makers_option"]

        # iterate until we are at the level where the 'key_data' can be found
        dict_for_iteration = self.outcomes.copy()
        formatted_data = pd.DataFrame(index=range(self.options))
        while dim_level > 1:
            dim_values = []
            tmp_dict = {}
            for key, value in dict_for_iteration.items():
                dim_values.append(key)
                # ensure keys are unique, so we don't lose any values
                tmp_dict.update({f"@{key}@{tmp_key}": tmp_value for tmp_key, tmp_value in value.items()})

            # for higher dimensions entries will be duplicated in the columns
            replicate_n = int(self.options / len(dim_values))
            # using a re(gex) expression the entries to ensure uniqueness in the for-loop are removed again.
            formatted_data[dim_names.pop(0)] = [
                re.sub(r"@.*?@", "", value) for value in dim_values for _ in range(replicate_n)
            ]
            dict_for_iteration = tmp_dict
            dim_level -= 1

        # add the values from 'key_data' | this can be either a list of dictionaries or a list of values
        key_data_list = get_values_from_target(self.outcomes, key_data)
        key_data_list_content = check_list_content(key_data_list)

        if key_data_list_content == "numeric":
            # if the final dimension is numeric we initialised too many rows
            formatted_data = formatted_data.drop_duplicates(ignore_index=True)
            formatted_data["value"] = key_data_list
        elif key_data_list_content == "dictionaries":
            formatted_data[key_data] = [key for dictionary in key_data_list for key, _ in dictionary.items()]
            formatted_data["value"] = [value for dictionary in key_data_list for _, value in dictionary.items()]
        return formatted_data

    @staticmethod
    def _apply_filters(dataframe: pd.DataFrame, drop_used: bool = False, **kwargs) -> pd.DataFrame and str:
        """
        This function applies filters, based on **kwargs arguments, on the dataframe and generates a corresponding
        name for the visual.
        :param dataframe: dataframe that needs filtering
        :param drop_used: indicator whether the used filter column should be dropped
        :return: a filtered dataframe with corresponding name
        """
        name_str = ""
        for arg, value in kwargs.items():
            if arg not in dataframe.columns:
                continue
            dataframe = dataframe[dataframe[arg] == value]
            if drop_used:
                dataframe = dataframe.drop(columns=[arg], axis=1)
            name_str += f" | {value}"

        # Check if dataframe is empty after applying filters
        if dataframe.empty:
            raise VisualizationError("No data for given selection. Are your arguments correct?")

        return dataframe, name_str

    def _create_table(self, key: str, **kwargs) -> pd.DataFrame.style:
        """
        This function creates a 2- or 3-dimensional table depending on the key.
        :param key: key of the values for the table
        :return: a styled table
        """
        table_data = self._format_data_for_visual(key)

        # Filter the data based on potentially provided arguments by the user.
        table_data, name_str = self._apply_filters(table_data, **kwargs)
        table_data = (
            table_data.set_index(["scenario", key])
            .pivot(columns="decision_makers_option", values="value")
            .rename_axis((None, None))
            .rename_axis(None, axis=1)
        )
        table_name = f"Values of {self._str_snake_case_to_text(key)}{name_str}"
        return self._table_styler(table_data.style, table_name)

    def _create_barchart(self, key: str, **kwargs) -> None:
        """
        This function creates and shows a barchart for a given data key.
        :param key: name of values of interest
        :return: a plotted barchart
        """
        dims = self._find_dimension_level(self.outcomes, key)
        if dims > 2 and "scenario" not in kwargs:
            raise VisualizationError(f"Too many dimensions ({dims}). Please specify a scenario")
        stacked = kwargs["stacked"] if "stacked" in kwargs else True
        show_legend = kwargs["show_legend"] if "show_legend" in kwargs else True

        appreciations = self._format_data_for_visual(key)
        bar_data, name_str = self._apply_filters(appreciations, drop_used=True, **kwargs)
        rest_cols = [col for col in bar_data.columns if col not in ["decision_makers_option", "value"]]
        bar_data = bar_data.pivot(index="decision_makers_option", columns=rest_cols, values="value").reset_index()
        axis = bar_data.plot.bar(x="decision_makers_option", stacked=stacked, color=self.colors, figsize=(10, 5))
        self._graph_styler(axis, f"Values of {self._str_snake_case_to_text(key)}{name_str}", show_legend)

        plt.show()

    def create_visual(self, visual_request: str, key: str, **kwargs):
        """
        This function redirects the visual_request based on the requested format to the correct helper function.
        Validation checks are performed on the available visuals and outputs.See 'available_outputs' and
        'available_kwargs' for all possible options.
        :param visual_request: type of visual that is requested
        :param key: key in output_dict containing the values to be visualised
        :param **kwargs: any additional arguments provide by the user
        :return: requested visual
        """
        self._validate_kwargs(**kwargs)
        if visual_request not in self.available_visuals:
            raise VisualizationError(f"'{visual_request}' is not a valid chart type")
        if key not in self.available_outputs:
            raise VisualizationError(f"'{key}' is not a valid option")

        return self.available_visuals[visual_request](key, **kwargs)



# pylint: disable=no-member

"""
This module contains the TRBS class. This is the parent class that deals with anything related to a Responsible
Business Simulator Case.
"""

from pathlib import Path
import os
import vlinder as vl
from vlinder.case_exporter import CaseExporter
from vlinder.case_importer import CaseImporter
from vlinder.evaluate import Evaluate
from vlinder.appreciate import Appreciate
from vlinder.visualize import Visualize


def list_demo_cases(file_path=None):
    """This function returns all demo cases that exist in the package"""
    file_path = file_path or Path(os.path.dirname(vl.__file__)) / "data"

    try:
        case_names = [name for name in os.listdir(file_path) if (file_path / name).is_dir()]
        return case_names
    except FileNotFoundError:
        raise FileNotFoundError(f"The directory {file_path} does not exist.")


class TheResponsibleBusinessSimulator:
    """
    This class is the base class of an tRBS-case and contains all necessary information to import data, evaluate
    dependencies and calculate appreciations.
    """

    def __init__(self, name, file_path=None, file_extension=None):
        self.file_path = file_path if file_path is not None else Path(os.path.dirname(vl.__file__)) / "data"
        self.file_extension = file_extension if file_extension is not None else "xlsx"
        self.name = name
        self.input_dict = {}
        self.dataframe_dict = {}
        self.output_dict = {}
        self.visualizer = None
        self.exporter = None

    def __str__(self):
        input_data_formatted = (
            "\n\n".join(f"{key}\n\t{value}" for key, value in self.input_dict.items())
            if self.input_dict
            else "First .build() a case to import data"
        )
        return (
            f"Case: {self.name} ({self.file_extension}) \n"
            f"Data location: {self.file_path} \n"
            f"Input data: \n {input_data_formatted}"
        )

    def _get_options(self):
        """
        This function calculates the amount of different options (or calculations) of the model:
        Amount of scenarios x Amount of decision makers options x Amount of key outputs
        """
        return (
            len(self.input_dict["scenarios"])
            * len(self.input_dict["decision_makers_options"])
            * len(self.input_dict["key_outputs"])
        )

    def build(self):
        """This function builds all necessary elements for a generic RBS case"""
        print(f"Creating '{self.name}'")
        case_import = CaseImporter(self.file_path, self.name, self.file_extension)
        self.input_dict, self.dataframe_dict = case_import.import_case()

    # TODO: param what to evaluate: single scenario = None, single dmo = None
    def evaluate(self):
        """This function deals with the evaluation of all dependencies"""
        case_evaluation = Evaluate(self.input_dict)
        self.output_dict = case_evaluation.evaluate_all_scenarios()

    def appreciate(self):
        """This function deals with the appreciation of the outcomes"""
        case_appreciation = Appreciate(self.input_dict, self.output_dict)
        case_appreciation.appreciate_all_scenarios()

    def visualize(self, visual_request, key, **kwargs):
        """This function deals with the visualizations of the outcomes"""
        # Set a Visualize class only if this has not yet been initialised.
        if not self.visualizer:
            self.visualizer = Visualize(self.output_dict, self._get_options())
        return self.visualizer.create_visual(visual_request, key, **kwargs)

    def transform(self, requested_format, output_path=None):
        """This function deals with transforming a case to a new format."""
        output_path = output_path if output_path is not None else Path.cwd() / "data"
        if not self.exporter:
            self.exporter = CaseExporter(output_path, self.name, self.dataframe_dict)
        self.exporter.create_template_for_requested_format(requested_format)



# Ignore PEP8 protected-access to client class | pylint: disable=W0212
"""
This module contains all tests for the Appreciate() class
"""
import pytest
from params import INPUT_DICT_BEERWISER, OUTPUT_DICT_BEERWISER
from vlinder.appreciate import Appreciate
from vlinder.utils import round_all_dict_values, get_values_from_target


@pytest.fixture(name="appreciate_beerwiser")
def fixture_appreciate_beerwiser():
    """
    This fixture initialises a Beerwiser case.
    :return: an Evaluate class for Beerwiser
    """
    return Appreciate(INPUT_DICT_BEERWISER, OUTPUT_DICT_BEERWISER)


def test_get_start_and_end_points(appreciate_beerwiser):
    """
    This function tests _get_start_and_end_points() to return a dictionary with a list containing the min. and max.
    found key output value, over all scenarios.
    :param appreciate_beerwiser:
    """
    result = appreciate_beerwiser._get_start_and_end_points()
    rounded_result = round_all_dict_values(result)
    expected_result = {
        "Accidents reduction": [3.49, 17.44],
        "Water use reduction": [1227272.73, 6818181.82],
        "Production cost reduction": [0.04, 0.05],
    }
    assert rounded_result == expected_result


@pytest.mark.parametrize(
    "value, args, expected_result",
    [
        (10, {"key_output": "Sample A", "key_output_smaller_the_better": 0, "key_output_linear": 1}, 33.33),
        (30, {"key_output": "Sample A", "key_output_smaller_the_better": 1, "key_output_linear": 1}, 0),
        (150, {"key_output": "Sample B", "key_output_smaller_the_better": 1, "key_output_linear": 0}, 15.71),
        (-20, {"key_output": "Sample A", "key_output_smaller_the_better": 0, "key_output_linear": 0}, 0),
        (10, {"key_output": "Sample C", "key_output_smaller_the_better": 0, "key_output_linear": 0}, 0),
    ],
)
def test_appreciate_single_key_output(appreciate_beerwiser, value, args, expected_result):
    """
    This function tests _appreciate_single_key_output to return the correct appreciation value for various combinations
    of arguments.
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    :param value: value of the key output
    :param args: arguments needed for appreciation of key output
    :param expected_result: expected appreciation value
    """
    # Set some own start and end points for testing purposes
    appreciate_beerwiser.start_and_end_points = {
        "Sample A": [5, 20],
        "Sample B": [0, 235],
        "Sample C": [10.345, 10.345],
    }
    result = round(appreciate_beerwiser._appreciate_single_key_output(value, args), 2)
    assert result == expected_result


def test_appreciate_single_decision_maker_option(appreciate_beerwiser):
    """
    This function tests aprreciate_single_decision_maker_option to return a dictionary that contains the correct values
    for key output, appreciations and weighted appreciations.
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """

    value_dict_in = appreciate_beerwiser.output_dict["Base case"]["Focus on training"]
    appreciate_beerwiser.appreciate_single_decision_maker_option(value_dict_in)

    result = round_all_dict_values(appreciate_beerwiser.output_dict["Base case"]["Focus on training"])
    expected_result = {
        "key_outputs": {
            "Accidents reduction": 17.44,
            "Water use reduction": 1336363.64,
            "Production cost reduction": 0.04,
        },
        "appreciations": {
            "Accidents reduction": 100.0,
            "Water use reduction": 1.95,
            "Production cost reduction": 56.73,
        },
        "weighted_appreciations": {
            "Accidents reduction": 33.33,
            "Water use reduction": 0.33,
            "Production cost reduction": 28.36,
        },
        "decision_makers_option_appreciation": 62.02,
    }
    assert result == expected_result


def test_appreciate_single_scenario_only_structure(appreciate_beerwiser):
    """
    This function tests the STRUCTURE after appreciate_single_scenario of the output dictionary. Only the keys and the
    length of the values are checked (should be 3 now) as these values are a result from
    appreciate_single_decision_maker_option() which is already checked above.
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """
    value_dict_in = appreciate_beerwiser.output_dict["Pessimistic"]
    appreciate_beerwiser.appreciate_single_scenario(value_dict_in)

    result = appreciate_beerwiser.output_dict["Pessimistic"]
    result_structure = {key: len(value) for key, value in result.items()}
    expected_structure = {"Equal spread": 4, "Focus on training": 4, "Focus on water recycling": 4}
    assert result_structure == expected_structure


def test_appreciate_all_scenarios_only_structure(appreciate_beerwiser):
    """
    This function tests for the presence of 'appreciations' and 'weighted_appreciations' in the output dictionary
    structure. The correctness of these values is already tested above.
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """
    appreciate_beerwiser.appreciate_all_scenarios()
    result = appreciate_beerwiser.output_dict

    # There should be 3 x 3 (dmo x scenario) appreciation dictionaries and 3 x 3 weighted appreciation dictionaries
    count_dictionaries = {
        "appreciations": len(get_values_from_target(result, "appreciations")),
        "weighted_appreciations": len(get_values_from_target(result, "weighted_appreciations")),
        "decision_makers_option_appreciation": len(
            get_values_from_target(result, "decision_makers_option_appreciation")
        ),
    }
    expected_count = {"appreciations": 9, "weighted_appreciations": 9, "decision_makers_option_appreciation": 9}
    assert count_dictionaries == expected_count


@pytest.mark.parametrize(
    "weights_list, expected_result",
    [([3, 5, 3, 15], 0.33), ([2, 4, 8, 20], 0.05), ([10, 8, 0, 2], 0), ([2, 3, 4, 0], 0)],
)
def test_apply_weights_single_key_output(appreciate_beerwiser, weights_list, expected_result):
    """
    This function tests _apply_weights_single_key_output to return the correct weight for a variety of
    different inputs. Also boundary cases with zero weights are tested.
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    :param weights_list: a list with weights needed to create weight dictionary
    :param expected_result: expected value for the weight
    """
    # prepare dictionary | easier to parametrize in this way
    weights_dict = {
        "key_output": weights_list[0],
        "theme": weights_list[1],
        "sum_within_theme": weights_list[2],
        "sum_theme": weights_list[3],
    }
    result = round(appreciate_beerwiser._apply_weights_single_key_output(weights_dict), 2)
    assert result == expected_result


def test_calculate_weights(appreciate_beerwiser):
    """
    This function tests _calculate_weights to return the correct values
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """
    result = appreciate_beerwiser._calculate_weights()
    rounded_result = [round(value, 2) for value in result]
    expected_result = [0.33, 0.17, 0.50]
    assert rounded_result == expected_result


def test_apply_weights(appreciate_beerwiser):
    """
    This function tests _apply_weights to return the correct weighted appreciations
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """
    appreciation_dict = {"Accidents reduction": 50, "Water use reduction": 49.756, "Production cost reduction": 81.12}
    result = appreciate_beerwiser._apply_weights(appreciation_dict)
    rounded_result = [round(value, 2) for value in result]
    expected_result = [16.67, 8.29, 40.56]
    assert expected_result == rounded_result


def test_apply_scenario_weights(appreciate_beerwiser):
    """
    This function tests _apply_scenario_weights to return the correct decision makers' option appreciations
    :param appreciate_beerwiser: an Appreciate() class for Beerwiser
    """
    appreciate_beerwiser._apply_scenario_weights()
    result = round_all_dict_values(appreciate_beerwiser.output_dict["Base case"]["Focus on training"])
    expected_result = {
        "key_outputs": {
            "Accidents reduction": 17.44,
            "Water use reduction": 1336363.64,
            "Production cost reduction": 0.04,
        },
        "appreciations": {
            "Accidents reduction": 100.0,
            "Water use reduction": 1.95,
            "Production cost reduction": 56.73,
        },
        "weighted_appreciations": {
            "Accidents reduction": 33.33,
            "Water use reduction": 0.33,
            "Production cost reduction": 28.36,
        },
        "decision_makers_option_appreciation": 62.02,
        "scenario_appreciations": 20.67,
    }
    assert result == expected_result



"""
This module contains some fixed parameters / values needed for the fixtures in the testing files.
"""
import numpy as np

# Input dictionary is created using repr(case.input_dict) on 17/07/2023 [Beerwiser].
INPUT_DICT_BEERWISER = {
    "configurations": np.array(["use_theme_weights"], dtype=object),
    "configuration_value": np.array([0]),
    "key_outputs": np.array(["Accidents reduction", "Water use reduction", "Production cost reduction"], dtype=object),
    "key_output_unit": np.array(["#/year", "hl/year", "%"], dtype=object),
    "key_output_theme": np.array(["People", "Planet", "Profit"], dtype=object),
    "themes": np.array(["People", "Planet", "Profit"], dtype="<U32"),
    "key_output_monetary": np.array([0, 0, 0]),
    "key_output_smaller_the_better": np.array([0, 0, 0]),
    "key_output_linear": np.array([1, 1, 0]),
    "key_output_automatic": np.array([1, 1, 1]),
    "key_output_start": np.array([np.nan, np.nan, np.nan]),
    "key_output_end": np.array([np.nan, np.nan, np.nan]),
    "key_output_threshold": np.array([np.nan, np.nan, np.nan]),
    "decision_makers_options": np.array(
        ["Equal spread", "Focus on training", "Focus on water recycling"], dtype=object
    ),
    "internal_variable_inputs": np.array(
        ["Invest in training of employees", "Invest in water recycling"], dtype=object
    ),
    "decision_makers_option_value": np.array([[150000, 150000], [250000, 50000], [50000, 250000]]),
    "scenarios": np.array(["Base case", "Optimistic", "Pessimistic"], dtype=object),
    "external_variable_inputs": np.array(["Cost of accident", "Effectiveness water recycling"], dtype=object),
    "scenario_value": np.array([[1.5e04, 9.8e-01], [1.2e04, 1.0e00], [2.0e04, 9.0e-01]]),
    "fixed_inputs": np.array(
        ["# employees", "Current # accidents", "Current production cost", "Current water use", "Water unit cost"],
        dtype=object,
    ),
    "fixed_input_value": np.array([5.0e02, 5.1e01, 7.5e06, 1.5e07, 5.0e-02]),
    "fixed_input_unit": np.array(["#", "#", "$", "hl", "$"], dtype=object),
    "intermediates": np.array(
        [
            "Accidents reduction %",
            "Water use reduction % when effective",
            "Water use reduction %",
            "Production cost reduction $",
            "Cost of training per employee",
            "New water use",
            "New # accidents",
        ],
        dtype=object,
    ),
    "intermediate_unit": np.array(["%", "%", "%", "$", "$", "hl/year", "#/year"], dtype=object),
    "intermediate_minimum": np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]),
    "intermediate_maximum": np.array([np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]),
    "dependencies": np.array([0, 1, 8, 9, 11, 2, 3, 4, 6, 10, 5, 12, 7]),
    "destination": np.array(
        [
            "Accidents reduction %",
            "Water use reduction % when effective",
            "Cost of training per employee",
            "New # accidents",
            "New water use",
            "Water use reduction %",
            "Accidents reduction",
            "Water use reduction",
            "Production cost reduction $",
            "New # accidents",
            "Production cost reduction $",
            "New water use",
            "Production cost reduction",
        ],
        dtype=object,
    ),
    "argument_1": np.array(
        [
            "Invest in training of employees",
            "Invest in water recycling",
            "Invest in training of employees",
            "Current # accidents",
            "Current water use",
            "Water use reduction % when effective",
            "Current # accidents",
            "Water use reduction %",
            "Accidents reduction",
            "Accidents reduction",
            "Water use reduction",
            "Water use reduction",
            "Production cost reduction $",
        ],
        dtype=object,
    ),
    "argument_2": np.array(
        [
            "",
            "",
            "# employees",
            "0",
            "0",
            "Effectiveness water recycling",
            "Accidents reduction %",
            "Current water use",
            "Cost of accident",
            "0",
            "Water unit cost",
            "0",
            "Current production cost",
        ],
        dtype=object,
    ),
    "operator": np.array(
        ["squeezed *", "squeezed *", "/", "*", "*", "*", "*", "*", "*", "-*", "*", "-*", "/"], dtype=object
    ),
    "maximum_effect": np.array(
        [0.48, 0.5, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    ),
    "accessibility": np.array(
        [0.95, 1.0, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    ),
    "probability_of_success": np.array(
        [0.9, 1.0, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    ),
    "saturation_point": np.array(
        [
            300000.0,
            275000.0,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
            np.nan,
        ]
    ),
    "theme_weight": np.array([2, 1, 3]),
    "key_output_weight": np.array([2.0, 1.0, 3.0]),
    "scenario_weight": np.array([2.0, 1.0, 3.0]),
}

# Generic output dictionary that is not related to a particular case
OUTPUT_DICT_GENERIC = {
    "SCEN A": {
        "DMO 1": {"val1": {"KO1": 2, "KO2": 4, "KO3": 3}, "val2": {"KO1": 0.4, "KO2": 0.4, "KO3": 0.2}, "val3": 5},
        "DMO 2": {"val1": {"KO1": 3, "KO2": 2, "KO3": 4}, "val2": {"KO1": 0.2, "KO2": 0.2, "KO3": 0.6}, "val3": 2},
        "DMO 3": {
            "val1": {"KO1": 5, "KO2": 2.5, "KO3": 3.5},
            "val2": {"KO1": 0.25, "KO2": 0.5, "KO3": 0.25},
            "val3": 4,
        },
    },
    "SCEN B": {
        "DMO 1": {"val1": {"KO1": 2.5, "KO2": 3, "KO3": 2}, "val2": {"KO1": 0.1, "KO2": 0.3, "KO3": 0.6}, "val3": 1.5},
        "DMO 2": {"val1": {"KO1": 4, "KO2": 1.5, "KO3": 2.5}, "val2": {"KO1": 0.2, "KO2": 0.2, "KO3": 0.6}, "val3": 2},
        "DMO 3": {"val1": {"KO1": 1, "KO2": 3, "KO3": 2.5}, "val2": {"KO1": 0.4, "KO2": 0.2, "KO3": 0.4}, "val3": 3.5},
    },
}


# Output dictionary is created using repr(case.output_dict) on 17/07/2023 [Beerwiser].
OUTPUT_DICT_BEERWISER = {
    "Base case": {
        "Equal spread": {
            "key_outputs": {
                "Accidents reduction": 10.4652,
                "Water use reduction": 4009090.909090909,
                "Production cost reduction": 0.047657672727272726,
            }
        },
        "Focus on training": {
            "key_outputs": {
                "Accidents reduction": 17.442,
                "Water use reduction": 1336363.6363636365,
                "Production cost reduction": 0.043793090909090907,
            }
        },
        "Focus on water recycling": {
            "key_outputs": {
                "Accidents reduction": 3.4883999999999995,
                "Water use reduction": 6681818.181818182,
                "Production cost reduction": 0.051522254545454546,
            }
        },
    },
    "Optimistic": {
        "Equal spread": {
            "key_outputs": {
                "Accidents reduction": 10.4652,
                "Water use reduction": 4090909.090909091,
                "Production cost reduction": 0.044017047272727275,
            }
        },
        "Focus on training": {
            "key_outputs": {
                "Accidents reduction": 17.442,
                "Water use reduction": 1363636.3636363638,
                "Production cost reduction": 0.03699810909090909,
            }
        },
        "Focus on water recycling": {
            "key_outputs": {
                "Accidents reduction": 3.4883999999999995,
                "Water use reduction": 6818181.818181818,
                "Production cost reduction": 0.051035985454545456,
            }
        },
    },
    "Pessimistic": {
        "Equal spread": {
            "key_outputs": {
                "Accidents reduction": 10.4652,
                "Water use reduction": 3681818.1818181816,
                "Production cost reduction": 0.052452654545454544,
            }
        },
        "Focus on training": {
            "key_outputs": {
                "Accidents reduction": 17.442,
                "Water use reduction": 1227272.7272727273,
                "Production cost reduction": 0.054693818181818184,
            }
        },
        "Focus on water recycling": {
            "key_outputs": {
                "Accidents reduction": 3.4883999999999995,
                "Water use reduction": 6136363.636363636,
                "Production cost reduction": 0.05021149090909091,
            }
        },
    },
}
